module InductiveExpr(
        Expression,
        Token,
        parse,
        tokenize,
        getToken,
        eval)

where

import Char


data Token = NoToken 
             | TokVal Int 
	     | TokSum 
	     | TokOpen 
	     | TokClose
         deriving (Eq,Read,Show)


getToken [] = (NoToken, [])
getToken (x:xs) = if isDigit x
                  then getNumber xs (digitToInt x)
                  else if isSumOp x
                  then (TokSum, xs)
                  else if isOpenPar x
                  then (TokOpen, xs)
                  else if isClosePar x
                  then (TokClose, xs)
                  else if isSpace x
                  then getToken xs
                  else error ("Token error at: "++(x:xs))

isSumOp x = x == '+'

isOpenPar x = x == '('

isClosePar x = x == ')'

getNumber [] n = (TokVal n, [])
getNumber (x:xs) n = if isDigit x
                      then getNumber xs ((n*10)+(digitToInt x))
                      else (TokVal n, (x:xs))


tokenize [] = []
tokenize (x:xs) = let
                   (nxttok,rest) = getToken (x:xs)
                  in if nxttok == NoToken
                     then []
                     else nxttok:(tokenize rest)


data Expression = Val Int | Sum Expression Expression
      deriving (Eq,Read,Show)


parse xs = let
             (expr,rest) = parseExpr xs
           in if null rest
              then expr
              else error (show rest)


parseExpr [] = error "a token sequence cannot represent an arithmetic expression"
parseExpr ((TokVal x):xs) = (Val x,xs)
parseExpr (TokOpen:xs) = let
	                   (ops,rest) = parseClose (parseSum xs)
                         in (opsToSum ops,rest)

parseClose (ops, TokClose:xs) = (ops, xs)
parseClose (ops, xs) = error(show (ops,xs))

opsToSum (op0, op1) = Sum op0 op1


parseSumOp0 :: [Token] -> (Expression,[Token])

parseSumOp1 :: (Expression,[Token]) -> ((Expression,Expression),[Token])

parseSumOptr :: (Expression,[Token]) -> (Expression,[Token])

parseSum = parseSumOp1 . parseSumOptr . parseSumOp0


parseSumOp0 = parseSumOperand
           
parseSumOp1 (op0,xs) = let
                     (op1,rest) = parseSumOperand xs
                    in ((op0,op1),rest)


parseSumOperand = parseExpr

parseSumOptr (expr,(TokSum:xs)) =  (expr, xs)
parseSumOptr xs = error (show xs)


treeEval (Val x) = x
treeEval (Sum x y) = (treeEval x) + (treeEval y) 

eval = treeEval . parse . tokenize
